/* 
 * File:   DAQChannelMapBase.h
 * Author: denis
 *
 * This is the base class for DAQChannelMap.
 * It is an abstract class.
 * Here we precalculate some stuff, which are going to be use in the derived class
 *
 * Created on February 10, 2012, 10:17 AM
 */

#ifndef _DAQCHANNELMAPBASE_H
#define	_DAQCHANNELMAPBASE_H

#include "DAQChannelMap/DAQChannelMapErrors.h"
#include "DAQChannelMap/DAQChannelMapConstants.h"

#include <string>
#include <vector>
#include <map>


namespace daqchannelmap {      //!< DAQ Channel Map Package

//#define BOUNDS_CHECK_ALL /* Define this macro to enable bounds checking */

/// This thing has to have a comma at the end or it could be entirely empty
#define CHAN_IDX_0_1    0x0,

    class DAQChannelMapBase {

            /// Some type defs
            typedef std::vector< uint32_t >                              numberOfPlanesInBlock_t;
            typedef std::vector< uint32_t >                              numberOfPlanesInDiblock_t;
            typedef std::map< DetPlane_TYPE, numberOfPlanesInBlock_t   > numberOfTypePlanesInBlock_t;
            typedef std::map< DetPlane_TYPE, numberOfPlanesInDiblock_t > numberOfTypePlanesInDiblock_t;
            typedef std::vector< std::vector <uint32_t> >                numberOfPlanesInBlockDetector_t;
            typedef std::vector< std::vector <uint32_t> >                numberOfDCMsInDiblockDetector_t;


        public:
            DAQChannelMapBase();
            virtual ~DAQChannelMapBase();

            /// Print Information about DAQ Channel
            virtual bool    print       (dchan daqchan    ) = 0;
            virtual bool    printOffline(lchan logicalchan) = 0;

            /// Get Detector Label from the enum
            std::string     getDetectorLabel(DetectorIdent) const;

            /// Pixel number from APDPixel
            pixel_t         APDPixel2Pixel(apd_pixel_t) const;
            /// APDPixel number from Pixel
            apd_pixel_t     pixel2APDPixel(pixel_t) const;

            /// Diblock type given detectorID and diblock number
            DiBlock_TYPE    getDiBlockTypeBase(DetectorIdent, diblock_t) const;
            /// Number of planes in the block given detectorID and block number
            uint32_t        getNumberOfPlanesInBlock  (DetectorIdent, block_t) const;
            /// Number of planes in the diblock given detectorID and diblock number
            uint32_t        getNumberOfPlanesInDiblock(DetectorIdent, diblock_t) const;
            /// Number of planes in the block given block type
            uint32_t        getNumberOfPlanesInBlock  (DetBlock_TYPE) const;
            /// Number of planes in the diblock given diblock type
            uint32_t        getNumberOfPlanesInDiblock(DiBlock_TYPE) const;
            /// Number of planes (X or Y) in the block given block type and type of plane(X or Y)
            uint32_t        getNumberOfPlaneTypeInBlock  (DetBlock_TYPE, DetPlane_TYPE);
            /// Number of planes (X or Y) in the diblock given diblock type and type of plane(X or Y)
            uint32_t        getNumberOfPlaneTypeInDiblock(DiBlock_TYPE , DetPlane_TYPE);

            /// Block type given diblock type an local block in diblock
            DetBlock_TYPE   getBlockTypeBase(DiBlock_TYPE, uint32_t local_block_number_in_diblock) const;
            /// Starting plane in the block given its type
            DetPlane_TYPE   getStartPlaneTypeInBlock(DetBlock_TYPE) const;


            /// The block number given the detectorID and offline plane number
            uint32_t        getBlockNumber            (DetectorIdent, offline_plane_t) const;
            /// The diblock number given the detectorID and offline plane number
            uint32_t        getDiblockNumber          (DetectorIdent, offline_plane_t) const;
            /// Local plane in block number given the detectorID and offline plane number
            offline_plane_t getLocalPlaneInBlockBase  (DetectorIdent, offline_plane_t) const;
            /// Local plane in diblock number given the detectorID and offline plane number
            offline_plane_t getLocalPlaneInDiblockBase(DetectorIdent, offline_plane_t) const;


            /// opposite plane view from the input. For instance Y = getOtherView(X)
            DetView_TYPE    getOtherView(DetView_TYPE) const;
            /// opposite plane view from the input. For instance Y = getOtherView(X)
            DetPlane_TYPE   getOtherPlaneType(DetPlane_TYPE) const;
            /// Conversion from DetPlane_TYPE to DetView_TYPE
            DetView_TYPE    detPlaneToDetView(DetPlane_TYPE) const;
            /// Conversion from DetView_TYPE to DetPlane_TYPE
            DetPlane_TYPE   detViewToDetPlane(DetView_TYPE) const;

            /// Number of DCMs for horizontal planes in the diblock given detectorID and Diblock type
            uint32_t        getNumberOfXDCM(DetectorIdent, DiBlock_TYPE) const;
            /// Number of DCMs for vertical planes in the diblock  given detectorID and Diblock type
            uint32_t        getNumberOfYDCM(DetectorIdent, DiBlock_TYPE) const;
            /// Total number of DCMs in the diblock  given detectorID and Diblock type
            uint32_t        getNumberOfDCM(DetectorIdent, DiBlock_TYPE) const;

            /// Number of channels in X plane for a given detectorID
            uint32_t        getNumberOfChannelsInXPlane(DetectorIdent) const;
            /// Number of channels in Y plane for a given detectorID
            uint32_t        getNumberOfChannelsInYPlane(DetectorIdent) const;

            /// Number of cells per DCM for a given diblock type
            uint32_t        getNumberOfCellsPerDCM(DiBlock_TYPE) const;

            /// Total number of blocks for a given detector
            uint32_t        getTotalNumberOfBlocks(DetectorIdent) const; ///< Total number of diblocks in the detector
            /// Total number of diblocks for a given detector
            uint32_t        getTotalNumberOfDiblocks(DetectorIdent) const; ///< Total number of diblocks in the detector

            // Error Message Related
            const char* errorString(DAQChanMapErrorCodes err_code) const;

        private:
            
            uint32_t        getBlockOrDiblock            (DetectorIdent,
                                                          offline_plane_t,
                                                          const numberOfPlanesInBlockDetector_t&) const;
            offline_plane_t getLocalPlaneInBlockOrDiblock(DetectorIdent,
                                                          offline_plane_t,
                                                          const numberOfPlanesInBlockDetector_t&) const;

            /// Initialization methods
            bool            initNumberOfCellsPerDCMInDiblock();
            bool            initBlockOrderInDiblock();
            bool            initPlanesInBlockID();
            bool            initNumberOfTypePlanesInBlock();
            bool            initPlanesInDiblockID();
            bool            initStartPlaneTypeInBlock();
            bool            initPlanesInBlockAndDiblockDetector();
            bool            initNumberOfDCMsInDiblockDetector();

            /// Checks for Initialization methods
            bool            isinitNumberOfCellsPerDCMInDiblock();
            bool            isinitBlockOrderInDiblock();
            bool            isinitPlanesInBlockID();
            bool            isinitNumberOfTypePlanesInBlock();
            bool            isinitPlanesInDiblockID();
            bool            isinitStartPlaneTypeInBlock();
            bool            isinitPlanesInBlockAndDiblockDetector();
            bool            isinitNumberOfDCMsInDiblockDetector();

            bool            initApdPixel2PixelMap();

        private:
                static const char*         DetectorLabels[];
                const static unsigned int  NUM_XCHAN_PLANE[];
                const static unsigned int  NUM_YCHAN_PLANE[];
                const static unsigned int  NUM_DIBLOCK_DET[];
                const static unsigned int  NUM_DCMS_DIBLOCK[];
                const static unsigned int  NUM_X_DCMS[];
                const static unsigned int  NUM_Y_DCMS[];
                const static          int* DetectorDiBlockConfig[];
                const static unsigned int  apd_pixel2cell_v0[],apd_cell2pixel_v0[];

		const static          int  DetectorDiBlockConfig_IPND[];
                const static          int  DetectorDiBlockConfig_NDOS[];
                const static          int  DetectorDiBlockConfig_NEARDET[];
                const static          int  DetectorDiBlockConfig_FARDET[];
                const static          int  DetectorDiBlockConfig_TESTBEAM_2Block_1x1[];
                const static          int  DetectorDiBlockConfig_TESTBEAM_3Block_1x1[];
                const static          int  DetectorDiBlockConfig_TESTBEAM_2Block_2x2[];
                const static          int  DetectorDiBlockConfig_TESTBEAM_3Block_2x2[];

                std::vector<uint32_t>           _planesInDiblockType; //Number of planes in diblock type
                std::vector<uint32_t>           _planesInBlockType;   //Number of planes in diblock type
                numberOfPlanesInBlockDetector_t _planesInDiblockDetector; // array depends on detector and diblock number
                numberOfPlanesInBlockDetector_t _planesInBlockDetector; // array depends on detector and diblock number
                numberOfPlanesInBlockDetector_t _BlockOrderInDiblock; // The order of blocks in the diblock

                std::vector<DetPlane_TYPE>      _startPlaneTypeInBlock;  ///< Does the block start with X or Y plane. Keyed by block

                numberOfTypePlanesInBlock_t     _numberOfTypePlanesInBlock;   /// Keyed bu the DetPlane_TYPE and
                numberOfTypePlanesInDiblock_t   _numberOfTypePlanesInDiBlock; /// Keyed bu the DetPlane_TYPE and

                std::vector<uint32_t>           _numberOfCellsPerDCMInDiblock; /// Number of cells that each DCM caries for each diblock type

                numberOfDCMsInDiblockDetector_t _numberOfXDCMsInDiblockDetector;// holds how many DCMs we make in a diblock of particular type and of particular detector
                numberOfDCMsInDiblockDetector_t _numberOfYDCMsInDiblockDetector;// holds how many DCMs we make in a diblock of particular type and of particular detector

                std::vector <pixel_t>           _apdPixel2CellMap;
                std::vector <apd_pixel_t>       _cell2ApdPixelMap;

    };
}// end of namespace daqchannelmap

#endif	/* _DAQCHANNELMAPBASE_H */

